from django import forms
from django.contrib.auth import authenticate
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django.conf import settings

from pybitid import bitid

from .models import Address
from .validators import validate_monero_address, validate_bitcoin_address
from .utils import generate_challenge


class ChallengeMixin(forms.Form):
    """
    Used on authentication forms to make sure an unique challenge is included.

    This mixin ensures that the challenge is always controlled by the server.
    """

    challenge = forms.CharField()

    def include_challange(self):
        new_challenge = bitid.build_uri(
            self.request.build_absolute_uri(), generate_challenge()
        )
        if not self.data:
            self.request.session["current_challenge"] = new_challenge
            self.initial["challenge"] = new_challenge

    def clean_challenge(self):
        challenge = self.cleaned_data.get("challenge")
        if not challenge or challenge != self.request.session.get("current_challenge"):
            raise forms.ValidationError(_("Invalid or outdated challenge"))

        return challenge


class SimpleLoginForm(ChallengeMixin, forms.Form):
    """Basic login form, that can be used as reference for implementation."""

    address = forms.CharField()
    signature = forms.CharField()

    error_messages = {
        "invalid_login": _("Please enter a correct Monero address or signature."),
        "inactive": _("This account is inactive."),
    }

    def __init__(self, request=None, *args, **kwargs):
        """When rendering the form (no data provided) a new challenge
        must be created."""
        super().__init__(*args, **kwargs)
        self.request = request
        self.user_cache = None
        self.include_challange()

    def clean(self):
        address = self.cleaned_data.get("address")
        challenge = self.cleaned_data.get("challenge")
        signature = self.cleaned_data.get("signature")

        if address and challenge and signature:
            self.user_cache = authenticate(
                self.request, address=address, challenge=challenge, signature=signature
            )
            if self.user_cache is None:
                raise self.get_invalid_login_error()
            else:
                self.confirm_login_allowed(self.user_cache)

        return self.cleaned_data

    def confirm_login_allowed(self, user):
        if not user.is_active:
            raise forms.ValidationError(
                self.error_messages["inactive"], code="inactive"
            )

    def get_user(self):
        return self.user_cache

    def get_invalid_login_error(self):
        return forms.ValidationError(
            self.error_messages["invalid_login"], code="invalid_login"
        )


class SimpleSignUpForm(ChallengeMixin, forms.Form):
    """Basic login form, that can be used as reference for implementation."""

    username = forms.CharField()
    address = forms.CharField()
    signature = forms.CharField()

    def __init__(self, request=None, *args, **kwargs):
        """When rendering the form (no data provided) a new challenge
        must be created."""
        super().__init__(*args, **kwargs)
        self.request = request
        self.include_challange()
        self.network = None

    def clean_address(self):
        self.network = None
        value = self.cleaned_data["address"]
        bitcoin_backend = "django_cryptolock.backends.BitcoinAddressBackend"
        monero_backend = "django_cryptolock.backends.MoneroAddressBackend"

        if bitcoin_backend in settings.AUTHENTICATION_BACKENDS:
            try:
                validate_bitcoin_address(value)
                self.network = Address.NETWORK_BITCOIN
            except ValidationError:
                pass

        if monero_backend in settings.AUTHENTICATION_BACKENDS:
            try:
                validate_monero_address(value)
                self.network = Address.NETWORK_MONERO
            except ValidationError:
                pass

        if not self.network:
            raise forms.ValidationError(_("Invalid address"))

        if Address.objects.filter(address=value).exists():
            raise forms.ValidationError(_("This address already exists"))

        return value

    def clean_username(self):
        value = self.cleaned_data["username"]
        if get_user_model().objects.filter(username=value).exists():
            raise forms.ValidationError(_("This username is already taken"))
        return value



import warnings
from secrets import token_hex

from django.conf import settings
from django.http.request import HttpRequest
from django.utils.translation import gettext_lazy as _

from monerorpc.authproxy import AuthServiceProxy
from pybitid import bitid


def verify_monero_signature(address: str, challenge: str, signature: str) -> bool:
    """Makes a request to wallet RPC to verify address and signature."""
    protocol = settings.DJCL_MONERO_WALLET_RPC_PROTOCOL
    host = settings.DJCL_MONERO_WALLET_RPC_HOST
    user = settings.DJCL_MONERO_WALLET_RPC_USER
    pwd = settings.DJCL_MONERO_WALLET_RPC_PASS
    wallet_rpc = AuthServiceProxy(f"{protocol}://{user}:{pwd}@{host}/json_rpc")

    result = wallet_rpc.verify(
        {"data": challenge, "address": address, "signature": signature}
    )

    return result.get("good", False)


def verify_bitcoin_signature(
    address: str, challenge: str, signature: str, request: HttpRequest
) -> bool:
    """Verifies if the provided bitcoin signature is valid."""
    network = getattr(settings, "DJCL_BITCOIN_NETWORK", None)
    if not network:
        warnings.warn(_("Please configure the bitcoin network in the settings file"))
    is_testnet = True if network == "testnet" else False
    callback_uri = request.build_absolute_uri()

    return bitid.challenge_valid(
        address, signature, challenge, callback_uri, is_testnet
    )


def generate_challenge():
    """Generates a new random challenge for the authentication."""
    return token_hex(8)



# Generated by Django 2.2.5 on 2020-02-18 19:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("django_cryptolock", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="address",
            name="network",
            field=models.PositiveSmallIntegerField(
                choices=[(1, "Monero"), (2, "Bitcoin")], default=1
            ),
        ),
        migrations.AlterField(
            model_name="address",
            name="address",
            field=models.CharField(max_length=106, unique=True),
        ),
    ]



from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django.conf import settings

from monero.address import Address
from pybitid.bitid import address_valid


def validate_monero_address(value):
    try:
        address = Address(value)
    except ValueError as e:
        raise ValidationError(
            _("%(value)s is not a valid address"), params={"value": value}
        )

    network = getattr(settings, "DJCL_MONERO_NETWORK", None)
    if not network:
        raise ValidationError(
            _("Please configure the monero network in the settings file")
        )
    if network == "mainnet" and not address.is_mainnet():
        raise ValidationError(_("Invalid address for mainnet"))
    elif network == "stagenet" and not address.is_stagenet():
        raise ValidationError(_("Invalid address for stagenet"))
    elif network == "testnet" and not address.is_testnet():
        raise ValidationError(_("Invalid address for testnet"))


def validate_bitcoin_address(value):
    network = getattr(settings, "DJCL_BITCOIN_NETWORK", None)
    if not network:
        raise ValidationError(
            _("Please configure the monero network in the settings file")
        )
    testnet = True if network == "testnet" else False
    if not address_valid(value, is_testnet=testnet):
        raise ValidationError(_(f"Invalid address for {network}"))



# -*- coding: utf-8 -*-
from django.conf.urls import url

from .views import CryptoLockLoginView, CryptoLockSignUpView


app_name = "django_cryptolock"
urlpatterns = [
    url(r"login", CryptoLockLoginView.as_view(), name="login"),
    url(r"signup", CryptoLockSignUpView.as_view(), name="signup"),
]



from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext_lazy as _

from monerorpc.authproxy import JSONRPCException

from .models import Address
from .utils import verify_monero_signature, verify_bitcoin_signature


class MoneroAddressBackend(ModelBackend):
    """Custom Monero-Cryptolock authentication backend."""

    def authenticate(
        self, request, address=None, challenge=None, signature=None, **kwargs
    ):
        """Validates the provided signature for the given address and challenge.

        This method currently relies on Wallet RPC access to verify the signature,
        in the future it should be done locally to be more reliable and more
        performant.
        """
        if not all([address, challenge, signature]):
            return None

        stored_address = (
            Address.objects.select_related("user")
            .filter(address=address, network=Address.NETWORK_MONERO)
            .first()
        )
        if not stored_address:
            return None
        try:
            is_valid = verify_monero_signature(
                stored_address.address, challenge, signature
            )
        except JSONRPCException:
            raise PermissionDenied(_("Error while validating signature"))

        if is_valid:
            return stored_address.user

        return None


class BitcoinAddressBackend(ModelBackend):
    """Custom Bitcoin-BitId authentication backend."""

    def authenticate(
        self, request, address=None, challenge=None, signature=None, **kwargs
    ):
        """
        Validates the provided signature for the given Bitcoin address and challenge.

        This method does not rely on any external components, everything is done locally.
        """
        if not all([address, challenge, signature]):
            return None

        stored_address = (
            Address.objects.select_related("user")
            .filter(address=address, network=Address.NETWORK_BITCOIN)
            .first()
        )
        if not stored_address:
            return None

        valid_signature = verify_bitcoin_signature(
            stored_address.address, challenge, signature, request
        )

        if valid_signature:
            return stored_address.user
        else:
            return None



# -*- coding: utf-8 -*-
from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError

from model_utils.models import TimeStampedModel

from .validators import validate_monero_address, validate_bitcoin_address


class Address(TimeStampedModel):
    """Addresses that belong to a given user account."""

    NETWORK_MONERO = 1
    NETWORK_BITCOIN = 2

    NETWORKS = ((NETWORK_MONERO, "Monero"), (NETWORK_BITCOIN, "Bitcoin"))

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    network = models.PositiveSmallIntegerField(choices=NETWORKS, default=NETWORK_MONERO)
    address = models.CharField(max_length=106, unique=True)

    class Meta:
        """Meta definition for Address."""

        verbose_name = _("Address")
        verbose_name_plural = _("Addresses")

    def __str__(self):
        """Unicode representation of Address."""
        return self.address

    def clean(self):
        try:
            if self.network == self.NETWORK_MONERO:
                validate_monero_address(self.address)
            else:
                validate_bitcoin_address(self.address)
        except ValidationError:
            raise ValidationError(_("Invalid address for the given network"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version(*file_paths):
    """Retrieves the version from django_cryptolock/__init__.py"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


version = get_version("django_cryptolock", "__init__.py")


if sys.argv[-1] == "publish":
    try:
        import wheel

        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    sys.exit()

if sys.argv[-1] == "tag":
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

readme = open("README.rst").read()
history = open("HISTORY.rst").read().replace(".. :changelog:", "")

with open("requirements.txt", "r") as f:
    requirements = [req.strip() for req in f.readlines()]

setup(
    name="django-cryptolock",
    version=version,
    description="""Django authentication using cryptocurrency wallets""",
    long_description=readme + "\n\n" + history,
    author="Gonçalo Valério",
    author_email="gon@ovalerio.net",
    url="https://github.com/dethos/django-cryptolock",
    packages=["django_cryptolock"],
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords="django-cryptolock",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
